/**
 SequenceMiner: multiple anomaly classifier
 
 @author: Eugene Turkov, {eturkov@email.arc.nasa.gov}
 @version: 1.0
 
 NASA Ames Research Center
 Computational Sciences Division
 Intelligent Data Understanding Group (IDU)
 Moffett Field, CA 94035
 */

package turkov.math.algorithms;
import turkov.util.TUSet;
import java.util.HashMap;

/**
    An instance of this class is a collection of data which also has the property
    of a set object. In addition to being a set, this object also has a property
    called "quality" which is the sum of the outlier scores of its elements divided
    by the total number of elements in this set.
 */ 
public class TUCluster extends TUSet
{          
    public TUCluster()
    {
        super();  
        this.quality = -1;
    }        
    
    /*
     Medoids are representative objects of a data set. The term used in computer
     computer science in fuzzy clustering algorithms. Basically the algorithm
     finds the center of a cluster and it takes the closest to the center (item)
     as "the medoid" upon which the distances to the other points are computed.
     
     @param the distance method to use to compute distances.
     */    
    private void calculateMedoid(String method) throws UnsupportedOperationException
    {
        //System.err.println("calculate medoid called...");
            this.quality = -1;
            double minSoFar = Double.MAX_VALUE;            
            validMedoid = true;            
            double[] distances = new double[this.size()];
            boolean finishedAllDistanceCalculations = true;
            
            // initial loop           
            for(int item = 0; item < this.size(); item++)
            {
                // inner loop to do the comparisons
                for(int otherItem = item; otherItem < this.size(); otherItem++)
                {
                    //System.err.println("CALCULATE MEDOID:  Comparing element " + item + " to " + otherItem);
                    Object[] a = (Object[])(((TUClusterableItem)this.get(item))
                    .getValue());
                    
                    Object[] b = (Object[])(((TUClusterableItem)this.get(otherItem))
                    .getValue());
                    
                    distances[item]
                            += TUDistanceMeasures.getDistance(a, b, method);
                    
                    // optimization
                    if(distances[item] > minSoFar && item > 0)
                    {
                        finishedAllDistanceCalculations = false;
                        break;
                    }                    
                }
                
                if(finishedAllDistanceCalculations)
                {
                    minSoFar = distances[item];
                    //this.quality = minSoFar;
                    // not setting quality here anymore, doing it outside now.
                    this.medoid = this.get(item);
                }
            }            
            // when this is done, medoid should be the actual medoid
    }        
    
    /*
     This method should be called to add more objects to this TUCluster object
     instance. Once an object is added, the current medoid is no longer valid and
     will eventually need to be recalculated if it is needed.
     
     @param object the object to add.
     */
    public void add(Object object)
    {
        validMedoid = false;
        super.add(object);
    }    
    
    /*
     This method should be called to add more objects to this TUCluster object
     instance. Once an object is added, the current medoid is no longer valid and
     will eventually need to be recalculated if it is needed.
     
     @param object the object to add.
     */
    public void add(Object object, boolean resetMedoid)
    {
        if(resetMedoid)
            add(object);
        
        else
            super.add(object);
    }      
    
    
    public void mergeContentsWith(TUSet foreignSet)
    {
        validMedoid = false;
        super.mergeContentsWith(foreignSet);
    }
    
    public void mergeWith(TUSet foreignSet)
    {
        validMedoid = false;
        super.mergeWith(foreignSet);
    }        

    /*
        "quality" is the sum of the outlier scores of the elements in this set
        divided by the total number of elements in this set. If the quality is
        either uninitialized or invalid, it will be recalculated.
     
        @param method the distance method used to calculate the distance between
        the elements of this cluster and its medoid.
     
        @return the quality of this cluster.
     */
    public double getQuality(String method)
    {
        if(this.size() == 0)
            return 0;
        
        if(this.quality < 0)
        {
            double sum = 0;
            for(int i = 0; i < this.size(); i++)
            {
                TUClusterableItem item = (TUClusterableItem)this.get(i);
                sum += item.getOutlierScore();
            }

            this.quality = sum/(double)this.size();
        }
        
        return quality;
    }
    
    /*
     Medoids are representative objects of a data set. The term used in computer
     computer science in fuzzy clustering algorithms. Basically the algorithm
     finds the center of a cluster and it takes the closest to the center (item)
     as "the medoid" upon which the distances to the other points are computed.
     If this cluster instance has either an invalid or uninitialized medoid, a new
     medoid will be calculated.
     
     @param method the distance method to use.
     
     @return the medoid of this cluster.
     */
    public Object getMedoid(String method)
    {
        if(!validMedoid || medoid == null)
            calculateMedoid(method);
        
        return medoid;
    }

    private Object medoid;    
    private boolean validMedoid;
    private double quality;
}
